@@ -2,6 +2,7 @@ from iam_syncr.errors import SyncrError, BadAmazon
 
 from boto.iam.connection import IAMConnection
 from contextlib import contextmanager
+from datadiff import diff
 import logging
 import urllib
 import json
@@ -28,8 +29,9 @@ class Amazon(object):
         if app_name not in useragent:
             sys.modules["boto.connection"].UserAgent = "{0} {1}/{2}".format(useragent, app_name, version)
 
-    def __init__(self, account_id, account_name, accounts):
+    def __init__(self, account_id, account_name, accounts, dry_run=False):
         self.changes = False
+        self.dry_run = dry_run
         self.accounts = accounts
         self.account_id = account_id
         self.account_name = account_name
@@ -78,6 +80,38 @@ class Amazon(object):
             else:
                 raise
 
+    @contextmanager
+    def ignore_boto_404(self):
+        """Ignore a BotoServerError 404"""
+        try:
+            yield
+        except boto.exception.BotoServerError as error:
+            if error.status == 404:
+                pass
+            else:
+                raise
+
+    def print_change(self, symbol, typ, changes=None, document=None, **kwargs):
+        """Print out a change"""
+        values = ", ".join("{0}={1}".format(key, val) for key, val in sorted(kwargs.items()))
+        print("{0} {1}({2})".format(symbol, typ, values))
+        if changes:
+            for change in changes:
+                print("\n".join("\t{0}".format(line) for line in change.split('\n')))
+        elif document:
+            print("\n".join("\t{0}".format(line) for line in document.split('\n')))
+
+    def change(self, symbol, typ, **kwargs):
+        """Print out a change and then do the change if not doing a dry run"""
+        self.print_change(symbol, typ, **kwargs)
+        if not self.dry_run:
+            try:
+                yield
+            except:
+                raise
+            else:
+                self.changes = True
+
     def split_role_name(self, name):
         """Split a role name into it's (name, path)"""
         split = name.split('/')
@@ -105,10 +139,12 @@ class Amazon(object):
 
     def info_for_profile(self, name):
         """Return what roles are attached to this profile if it exists"""
+        profiles = []
         role_name, _ = self.split_role_name(name)
-        with self.catch_boto_400("Couldn't list instance profiles associated with a role", role=role_name):
-            result = self.connection.list_instance_profiles_for_role(role_name)
-        profiles = result["list_instance_profiles_for_role_response"]["list_instance_profiles_for_role_result"]["instance_profiles"]
+        with self.ignore_boto_404():
+            with self.catch_boto_400("Couldn't list instance profiles associated with a role", role=role_name):
+                result = self.connection.list_instance_profiles_for_role(role_name)
+            profiles = result["list_instance_profiles_for_role_response"]["list_instance_profiles_for_role_result"]["instance_profiles"]
 
         existing_profile = None
         for profile in profiles:
@@ -128,10 +164,9 @@ class Amazon(object):
         existing_roles_in_profile = self.info_for_profile(role_name)
         if existing_roles_in_profile is None:
             try:
-                log.info("Making instance profile\tname=%s", role_name)
                 with self.catch_boto_400("Couldn't create instance profile", instance_profile=role_name):
-                    self.connection.create_instance_profile(role_name)
-                self.changes = True
+                    for _ in self.change("+", "instance_profile", profile=role_name):
+                        self.connection.create_instance_profile(role_name)
             except boto.exception.BotoServerError as error:
                 if error.status == 409 and error.code == "EntityAlreadyExists":
                     # I'd rather ignore this conflict, than list all the instance_profiles
@@ -142,37 +177,34 @@ class Amazon(object):
 
         if existing_roles_in_profile and any(rl != role_name for rl in existing_roles_in_profile):
             for role in [rl for rl in existing_roles_in_profile if rl != role_name]:
-                log.info("Removing role from existing instance_profile\tprofile=%s\trole=%s", role_name, role)
                 with self.catch_boto_400("Couldn't remove role from an instance profile", profile=role_name, role=role):
-                    self.connection.remove_role_from_instance_profile(role_name, role)
-                self.changes = True
+                    for _ in self.change("-", "instance_profile_role", profile=role_name, role=role):
+                        self.connection.remove_role_from_instance_profile(role_name, role)
 
         if not existing_roles_in_profile or not any(rl == role_name for rl in existing_roles_in_profile):
-            log.info("Adding role to an instance_profile\tprofile=%s\trole=%s", role_name, role_name)
             with self.catch_boto_400("Couldn't add role to an instance profile", role=role_name, instance_profile=role_name):
-                self.connection.add_role_to_instance_profile(role_name, role_name)
-            self.changes = True
+                for _ in self.change("+", "instance_profile_role", profile=role_name, role=role_name):
+                    self.connection.add_role_to_instance_profile(role_name, role_name)
 
     def create_role(self, name, trust_document, policies=None):
         """Create a role"""
         role_name, role_path = self.split_role_name(name)
         with self.catch_boto_400("Couldn't create role", "{0} trust document".format(name), trust_document, role=name):
-            log.info("Creating a new role\trole=%s", name)
-            self.connection.create_role(role_name, assume_role_policy_document=trust_document, path=role_path)
-            self.changes = True
+            for _ in self.change("+", "role", role=role_name, document=trust_document):
+                self.connection.create_role(role_name, assume_role_policy_document=trust_document, path=role_path)
 
         # And add our permissions
         if policies:
             for policy_name, document in policies.items():
                 if document:
                     with self.catch_boto_400("Couldn't add policy", "{0} - {1} Permission document".format(role_name, policy_name), document, role=role_name, policy_name=policy_name):
-                        self.connection.put_role_policy(role_name, policy_name, document)
-                        self.changes = True
+                        for _ in self.change("+", "role_policy", role=role_name, policy=policy_name, document=document):
+                            self.connection.put_role_policy(role_name, policy_name, document)
 
     def compare_trust_document(self, role_info, trust_document):
         """Say whether the provided trust document is the same as the one in the role_info"""
         if not role_info or not role_info.get("role", {}).get("assume_role_policy_document"):
-            return False
+            return []
 
         unquoted = urllib.unquote(role_info["role"]["assume_role_policy_document"])
         return self.compare_two_documents(unquoted, trust_document)
@@ -180,25 +212,53 @@ class Amazon(object):
     def compare_two_documents(self, doc1, doc2):
         """Compare two documents by converting them into json objects and back to strings and compare"""
         try:
-            first = json.dumps(json.loads(doc1), indent=2, sort_keys=True).strip()
+            first = json.loads(doc1)
         except (ValueError, TypeError):
-            return False
+            return
 
         try:
-            second = json.dumps(json.loads(doc2), indent=2, sort_keys=True).strip()
+            second = json.loads(doc2)
         except (ValueError, TypeError):
-            return False
+            return
 
-        return first == second
+        # Ordering the principals because the ordering amazon gives me hates me
+        def sort_statement(statement):
+            for principal in (statement.get("Principal", None), statement.get("NotPrincipal", None)):
+                if principal:
+                    for principal_type in ("AWS", "Federated", "Service"):
+                        if principal_type in principal and type(principal[principal_type]) is list:
+                            principal[principal_type] = sorted(principal[principal_type])
+        def sort_key(statement, key):
+            if key in statement and type(statement[key]) is list:
+                statement[key] = sorted(statement[key])
+        for document in (first, second):
+            if "Statement" in document:
+                if type(document["Statement"]) is dict:
+                    sort_statement(document["Statement"])
+                    sort_key(document["Statement"], "Resource")
+                    sort_key(document["Statement"], "NotResource")
+                else:
+                    for statement in document["Statement"]:
+                        sort_statement(statement)
+                        sort_key(statement, "Resource")
+                        sort_key(statement, "NotResource")
+
+        difference = diff(first, second, fromfile="current", tofile="new").stringify()
+        if difference:
+            lines = difference.split('\n')
+            if any(line.strip().startswith("@@") and line.strip().endswith("@@") for line in lines):
+                for line in lines:
+                    yield line
 
     def modify_role(self, role_info, name, trust_document, policies=LeaveAlone):
         """Modify a role"""
         role_name, _ = self.split_role_name(name)
-        if trust_document and not self.compare_trust_document(role_info, trust_document):
-            with self.catch_boto_400("Couldn't modify trust document", "{0} assume document".format(role_name), trust_document, role=role_name):
-                log.info("Modifying trust document\trole=%s", role_name)
-                self.connection.update_assume_role_policy(role_name, trust_document)
-                self.changes = True
+        if trust_document:
+            changes = list(self.compare_trust_document(role_info, trust_document))
+            if changes:
+                with self.catch_boto_400("Couldn't modify trust document", "{0} assume document".format(role_name), trust_document, role=role_name):
+                    for _ in self.change("M", "trust_document", role=role_name, changes=changes):
+                        self.connection.update_assume_role_policy(role_name, trust_document)
 
         if policies is LeaveAlone:
             return
@@ -214,20 +274,22 @@ class Amazon(object):
             log.info("Role has unknown policies that will be disassociated\trole=%s\tunknown=%s", role_name, unknown)
             for policy in unknown:
                 with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=role_name):
-                    self.connection.delete_role_policy(role_name, policy)
-                self.changes = True
+                    for _ in self.change("-", "role_policy", role=role_name, policy=policy):
+                        self.connection.delete_role_policy(role_name, policy)
 
         for policy, document in policies.items():
             if not document:
                 if policy in current_policies:
-                    log.info("Removing policy\trole=%s\tpolicy=%s", role_name, policy)
                     with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=role_name):
-                        self.connection.delete_role_policy(role_name, policy)
-                    self.changes = True
+                        for _ in self.change("-", "policy", role=role_name, policy=policy):
+                            self.connection.delete_role_policy(role_name, policy)
             else:
                 needed = False
+                changes = None
+
                 if policy in current_policies:
-                    if not self.compare_two_documents(current_policies.get(policy), document):
+                    changes = list(self.compare_two_documents(current_policies.get(policy), document))
+                    if changes:
                         log.info("Overriding existing policy\trole=%s\tpolicy=%s", role_name, policy)
                         needed = True
                 else:
@@ -236,10 +298,9 @@ class Amazon(object):
 
                 if needed:
                     with self.catch_boto_400("Couldn't add policy document", "{0} - {1} policy document".format(role_name, policy), document, role=role_name, policy=policy):
-                        self.connection.put_role_policy(role_name, policy, document)
-                        log.debug(policy)
-                        log.debug(document)
-                        log.debug('------')
+                        symbol = "M" if changes else "+"
+                        for _ in self.change(symbol, "role_policy", role=role_name, policy=policy, changes=changes, document=document):
+                            self.connection.put_role_policy(role_name, policy, document)
 
     def current_role_policies(self, name, comparing):
         """Get the current policies for some role"""
@@ -262,10 +323,9 @@ class Amazon(object):
         """Remove the role if it exists"""
         role_name, _ = self.split_role_name(name)
         if self.has_role(role_name):
-            log.info("Deleting role\trole=%s", role_name)
             with self.catch_boto_400("Couldn't delete a role", role=role_name):
-                self.connection.delete_role(role_name)
-            self.changes = True
+                for _ in self.change("-", "role", role=role_name):
+                    self.connection.delete_role(role_name)
         else:
             log.info("Role already deleted\trole=%s", role_name)
 
